#ifndef _IP_CONNTRACK_TUPLE_H
#define _IP_CONNTRACK_TUPLE_H

#include <linux/types.h>
#include <linux/netfilter/nf_conntrack_tuple_common.h>

/* A `tuple' is a structure containing the information to uniquely
  identify a connection.  ie. if two packets have the same tuple, they
  are in the same connection; if not, they are not.

  We divide the structure along "manipulatable" and
  "non-manipulatable" lines, for the benefit of the NAT code.
*/

/* The protocol-specific manipulable parts of the tuple: always in
   network order! */
union ip_conntrack_manip_proto
{
	/* Add other protocols here. */
	u_int16_t all;

	struct {
		__be16 port;
	} tcp;
	struct {
		u_int16_t port;
	} udp;
	struct {
		u_int16_t id;
	} icmp;
	struct {
		u_int16_t port;
	} sctp;
	struct {
		__be16 key;	/* key is 32bit, pptp only uses 16 */
	} gre;
};

/* The manipulable part of the tuple. */
struct ip_conntrack_manip
{
	u_int32_t ip;
	union ip_conntrack_manip_proto u;
#if 1 // CONFIG_SCHIZOPHRENIC_ROUTING
	// This is set if the source should be accepted from any port.
	u_int8_t ignoreport;

	// This is the pointer to the source interface (used in SNAT), or NULL if
	// no such given or is unknown.
	struct net_device *dev;
#endif

#if 1 // CONFIG_CONNTRACK_TOS
	//
	int tos;
#endif
};

/* This contains the information to distinguish a connection. */
struct ip_conntrack_tuple
{
	struct ip_conntrack_manip src;

	/* These are the parts of the tuple which are fixed. */
	struct {
		u_int32_t ip;
		union {
			/* Add other protocols here. */
			u_int16_t all;

			struct {
				u_int16_t port;
			} tcp;
			struct {
				u_int16_t port;
			} udp;
			struct {
				u_int8_t type, code;
			} icmp;
			struct {
				u_int16_t port;
			} sctp;
			struct {
				__be16 key;	/* key is 32bit, 
						 * pptp only uses 16 */
			} gre;
		} u;

		/* The protocol. */
		u_int8_t protonum;

		/* The direction (for tuplehash) */
		u_int8_t dir;
#if 1 // CONFIG_SCHIZOPHRENIC_ROUTING
		/* The destination device pointer (used in DNAT), or NULL
		   if no such given or is unknown. */
		struct net_device *dev;
#endif

#if 1 // CONFIG_CONNTRACK_TOS
		//
		int tos;
#endif
	} dst;
};

/* This is optimized opposed to a memset of the whole structure.  Everything we
 * really care about is the  source/destination unions */
#define IP_CT_TUPLE_U_BLANK(tuple) 				\
	do {							\
		(tuple)->src.u.all = 0;				\
		(tuple)->dst.u.all = 0;				\
	} while (0)

#ifdef __KERNEL__

#define DUMP_TUPLE(tp)						\
DEBUGP("tuple %p: %u %u.%u.%u.%u:%hu -> %u.%u.%u.%u:%hu\n",	\
       (tp), (tp)->dst.protonum,				\
       NIPQUAD((tp)->src.ip), ntohs((tp)->src.u.all),		\
       NIPQUAD((tp)->dst.ip), ntohs((tp)->dst.u.all))

/* If we're the first tuple, it's the original dir. */
#define DIRECTION(h) ((enum ip_conntrack_dir)(h)->tuple.dst.dir)

/* Connections have two entries in the hash table: one for each way */
struct ip_conntrack_tuple_hash
{
	struct list_head list;

	struct ip_conntrack_tuple tuple;
};

#endif /* __KERNEL__ */

static inline int ip_ct_tuple_srcdev_equal(const struct ip_conntrack_tuple *t1,
                                           const struct ip_conntrack_tuple *t2)
{
#if 1 // CONFIG_SCHIZOPHRENIC_ROUTING
	return t1->src.dev == t2->src.dev ||
	       // If the dievice is not given, evaluate it as equal:
	       t1->src.dev == NULL ||
	       t2->src.dev == NULL;
#else
	return 1;
#endif
}

static inline int ip_ct_tuple_dstdev_equal(const struct ip_conntrack_tuple *t1,
                                           const struct ip_conntrack_tuple *t2)
{
#if 1 // CONFIG_SCHIZOPHRENIC_ROUTING
	return t1->dst.dev == t2->dst.dev ||
	       // If the dievice is not given, evaluate it as equal:
	       t1->dst.dev == NULL ||
	       t2->dst.dev == NULL;
#else
	return 1;
#endif
}

static inline int ip_ct_tuple_src_equal(const struct ip_conntrack_tuple *t1,
				        const struct ip_conntrack_tuple *t2)
{
	return t1->src.ip == t2->src.ip
		&& t1->src.u.all == t2->src.u.all;
}

static inline int ip_ct_tuple_dst_equal(const struct ip_conntrack_tuple *t1,
				        const struct ip_conntrack_tuple *t2)
{
	return t1->dst.ip == t2->dst.ip
		&& t1->dst.u.all == t2->dst.u.all
		&& t1->dst.protonum == t2->dst.protonum;
}

static inline int ip_ct_tuple_equal(const struct ip_conntrack_tuple *t1,
				    const struct ip_conntrack_tuple *t2)
{
	return ip_ct_tuple_src_equal(t1, t2) && ip_ct_tuple_dst_equal(t1, t2);
}

static inline int ip_ct_tuple_equal_nosrcport(const struct ip_conntrack_tuple *t1,
				              const struct ip_conntrack_tuple *t2)
{
	return t1->src.ip == t2->src.ip
	       && (t2->src.ignoreport || t1->src.u.all == t2->src.u.all)
	       && ip_ct_tuple_dst_equal(t1, t2);
}

static inline int ip_ct_tuple_equal_srcdev(const struct ip_conntrack_tuple *t1,
                                           const struct ip_conntrack_tuple *t2)
{
	return ip_ct_tuple_equal(t1, t2)
	       && ip_ct_tuple_srcdev_equal(t1, t2);
}

static inline int ip_ct_tuple_equal_special(const struct ip_conntrack_tuple *t1,
                                            const struct ip_conntrack_tuple *t2)
{
	return ip_ct_tuple_equal_nosrcport(t1, t2)
	       && ip_ct_tuple_srcdev_equal(t1, t2);
}

static inline int ip_ct_tuple_equal_dstdev(const struct ip_conntrack_tuple *t1,
                                           const struct ip_conntrack_tuple *t2)
{
	return ip_ct_tuple_equal(t1, t2)
	       && ip_ct_tuple_dstdev_equal(t1, t2);
}

static inline int ip_ct_tuple_mask_cmp(const struct ip_conntrack_tuple *t,
                                       const struct ip_conntrack_tuple *tuple,
                                       const struct ip_conntrack_tuple *mask)
{
	return ip_ct_tuple_srcdev_equal(t, tuple)
		&& ip_ct_tuple_dstdev_equal(t, tuple)
	    && !(((t->src.ip ^ tuple->src.ip) & mask->src.ip)
	      || ((t->dst.ip ^ tuple->dst.ip) & mask->dst.ip)
	      || ((t->src.u.all ^ tuple->src.u.all) & mask->src.u.all)
	      || ((t->dst.u.all ^ tuple->dst.u.all) & mask->dst.u.all)
	      || ((t->dst.protonum ^ tuple->dst.protonum)
	          & mask->dst.protonum));
}

#endif /* _IP_CONNTRACK_TUPLE_H */
